Index: FL/Fl_Widget.H
===================================================================
--- FL/Fl_Widget.H	(revision 12937)
+++ FL/Fl_Widget.H	(working copy)
@@ -842,6 +842,9 @@
    */
   unsigned int visible_focus() { return flags_ & VISIBLE_FOCUS; }
 
+  static unsigned int fribidi();
+  static void fribidi(int v);
+
   /** The default callback for all widgets that don't set a callback.
 
     This callback function puts a pointer to the widget on the queue
Index: src/Fl_Widget.cxx
===================================================================
--- src/Fl_Widget.cxx	(revision 12937)
+++ src/Fl_Widget.cxx	(working copy)
@@ -24,7 +24,66 @@
 #include <stdlib.h>
 #include "flstring.h"
 
+#if defined(FLTK_USE_FRIBIDI)
 
+#include <fribidi.h>
+#define FRIBIDI_MAX_STRLEN 65000
+
+static unsigned int fribidi_ = 1;
+
+unsigned int Fl_Widget::fribidi(void) {
+  return fribidi_;
+}
+
+void Fl_Widget::fribidi(int v) {
+  fribidi_ = v ? 1 : 0;
+}
+
+/** Returns pointer to allocated string or NULL on error. */
+static char *fribidi_parse_line(const char *input)
+{
+  char buffer[FRIBIDI_MAX_STRLEN];
+  size_t size;
+  FriBidiParType base = FRIBIDI_PAR_LTR;
+  FriBidiStrIndex len, *ltov = NULL, *vtol = NULL;
+  FriBidiChar logical[FRIBIDI_MAX_STRLEN];
+  FriBidiChar visual[FRIBIDI_MAX_STRLEN];
+  FriBidiLevel *levels = NULL;
+  FriBidiCharSet charset;
+
+  if (!input || (size = strlen(input)) == 0)
+    return NULL;
+
+  charset = fribidi_parse_charset("UTF-8");
+  fribidi_set_mirroring(true);
+  fribidi_set_reorder_nsm(false);
+
+  strncpy(buffer, input, size);
+  len = fribidi_charset_to_unicode(charset, buffer, size, logical);
+
+  if (len == 0)
+    return NULL;
+
+  if (fribidi_log2vis(logical, len, &base, visual, ltov, vtol, levels)) {
+    len = fribidi_remove_bidi_marks(visual, len, ltov, vtol, levels);
+    if (len > 0) {
+      fribidi_unicode_to_charset(charset, visual, len, buffer);
+      return strdup(buffer);
+    }
+  }
+
+  return NULL;
+}
+
+#else
+
+/* no fribidi support */
+unsigned int Fl_Widget::fribidi(void) { return 0; }
+void Fl_Widget::fribidi(int) { }
+
+#endif
+
+
 ////////////////////////////////////////////////////////////////
 // for compatibility with Forms, all widgets without callbacks are
 // inserted into a "queue" when they are activated, and the forms
@@ -111,7 +170,17 @@
 
   x_ = X; y_ = Y; w_ = W; h_ = H;
 
-  label_.value	 = L;
+#if defined(FLTK_USE_FRIBIDI)
+  char *str;
+  if (L && fribidi() && (str = fribidi_parse_line(L))) {
+    label_.value = str;
+    set_flag(COPIED_LABEL);
+  } else {
+    label_.value = L;
+  }
+#else
+  label_.value   = L;
+#endif
   label_.image   = 0;
   label_.deimage = 0;
   label_.type	 = FL_NORMAL_LABEL;
@@ -274,7 +343,6 @@
   return 0;
 }
 
-
 void Fl_Widget::label(const char *a) {
   if (flags() & COPIED_LABEL) {
     // reassigning a copied label remains the same copied label
@@ -283,11 +351,21 @@
     free((void *)(label_.value));
     clear_flag(COPIED_LABEL);
   }
+
+#if defined(FLTK_USE_FRIBIDI)
+  char *b = NULL;
+  if (fribidi() && (b = fribidi_parse_line(a))) {
+    label_.value=b;
+    set_flag(COPIED_LABEL);
+  } else {
+    label_.value=a;
+  }
+#else
   label_.value=a;
+#endif
   redraw_label();
 }
 
-
 void Fl_Widget::copy_label(const char *a) {
   // reassigning a copied label remains the same copied label
   if ((flags() & COPIED_LABEL) && (label_.value == a))
Index: src/Fl_Window.cxx
===================================================================
--- src/Fl_Window.cxx	(revision 12937)
+++ src/Fl_Window.cxx	(working copy)
@@ -519,9 +519,17 @@
 }
 
 void Fl_Window::label(const char *name, const char *mininame) {
+# if defined(FLTK_USE_FRIBIDI)
+  // right-to-left labels are handled by the window manager
+  int v = Fl_Widget::fribidi();
+  Fl_Widget::fribidi(0);
+# endif
   Fl_Widget::label(name);
   iconlabel_ = mininame;
   pWindowDriver->label(name, mininame);
+# if defined(FLTK_USE_FRIBIDI)
+  Fl_Widget::fribidi(v);
+# endif
 }
 
 void Fl_Window::show() {
